/*
** Module   : tty.screen.js
** Abstract : screen management.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20140110 First version.
** 002 MAG 20140204 Implements text attributes.
** 003 MAG 20140211 Implements COLOR.
** 004 MAG 20140217 Minor changes.
** 005 MAG 20140219 Performances improvements.
*/

tty.screen = (function()
{
   /* exports object */
   var my = {};
   
   /* used to calculate character width */
   var calibration = 'THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG';

   /* canvas */
   var canvas;
   
   /* canvas 2d context */
   var ctx;
   
   /* canvas font */
   var font, bold;
   
   /* screen dimensions */
   var rows, cols;      
   
   /* font width and height */
   var dx, dy;

   /* color palette */
   my.palette;
   
   /*
   ** Module   : video
   ** Abstract : screen video memory
   */
   var video = (function()
   {
      /* exports object */
      var my = {};      
      
      /* video memory */
      var memo;      
      
      /* initialize video memory */
      my.init = function(size, color)
      {
         my.color = color;         
         memo = new Array(size.rows);
         for (i = 0;i < size.rows;i++) 
         {
            memo[i] = new Array(size.cols);
            for (j = 0;j < size.cols;j++)
            {
               var cell = {'c':' ','f':'white','b':'black','a':0,'k':0};
               cell.f = color.f;
               cell.b = color.b;
               memo[i][j] = cell;
            }
         }
      };

      /* clear video memory */
      my.clear = function(rows, cols, color)
      {
         for (i = 0;i < rows;i++) 
         {
            for (j = 0;j < cols;j++)
            {
               var v = memo[i][j]; 
               v.c = ' ';
               v.f = color.f;
               v.b = color.b;
               v.a = 0;
               v.k = 0;
            }
         }
      };

      /* get cell value */
      my.get = function(row, col)
      {
         return memo[row][col];
      };

      /* set cell value*/
      my.put = function(c, row, col, a, k)
      {
         var v = memo[row][col]; 
         v.c = c;
         v.a = a;
         v.k = k;
      };
      
      /* exports */
      return my;
   })();

   /*
   ** Module   : cursor 
   ** Abstract : screen cursor management
   */
   var cursor = (function(screen)
   {
      /* exports object */
      var my = {};
      
      /* hide cursor */
      my.hide = function()
      {
         screen.drawCursor(false);
      };
      
      /* show cursor */
      my.show = function()
      {
         screen.drawCursor(my.visible);
      };

      /* set cursor position */
      my.setPosition = function(row, col)
      {
         my.hide();
         my.row = row;
         my.col = col;
         my.show();
      };
      
      /*
      ** Module   : blinking 
      ** Abstract : screen cursor blinking management
      */
      my.blinking = (function(cursor, screen) 
      {
         /* exports object */
         var my = {};
         
         /* timer */
         var timer;

         /* stop blinking */
         my.off = function()
         {
            window.clearInterval(timer);
         };

         /* start blinking */
         my.on = function()
         {
            var blink = true;
            timer = window.setInterval(function()
            {
               screen.drawCursor(blink & cursor.visible);
               blink = !blink;
            }, 500);
         };

         /* exports */
         return my;
      }(my, screen));
      
      /* initialize cursor */
      my.init = function(cursor)
      {
         my.row = 0;
         my.col = 0;
         my.type = cursor.type;
         my.visible = cursor.visible;
         if (cursor.blinking) 
         {
            my.blinking.on();
         }
      };
      
      /* exports */
      return my;
   })(my);
   
   /* draw cursor */
   my.drawCursor = function(on)
   {
      var v = video.get(cursor.row, cursor.col);
      
      var p = getColors(v.k);
                       
      var x = cursor.col * dx;
      var y = cursor.row * dy;
      
      if (on)
      {
         if (cursor.type === 'solid')
         {         
            p.f = video.color.b;
            p.b = video.color.f;
            drawText(x, y, v.c, 0, p);
         } 
         else 
         {
            drawText(x, y, v.c, v.a, p);         
            // draw line cursor           
            ctx.fillStyle = video.color.f;
            ctx.fillRect(x, y + dy - 4, dx, 4);
         }
      }
      else
      {
         drawText(x, y, v.c, v.a, p);         
      }
   };
   
   /* print an error message on the last line */
   my.error = function(text)
   {
      var row = rows-1;
      var col = 0;
      
      var x = col * dx;
      var y = row * dy;
      
      var l = text.length;
      var max = dx * l;
      
      var v = {};
      v.f = video.color.f;      
      v.b = video.color.b;      
      drawText(x, y, text, 0, v);
      
      tty.sound.beep();      
   };
   
   /* initialize canvas */
   my.init = function(cfg)
   {
      if (cfg.font.name === undefined || cfg.font.name.toLowerCase() === 'monospaced')
      {
         cfg.font.name = 'monospace';
      }
      rows = cfg.canvas.size.rows;
      cols = cfg.canvas.size.cols;    
      video.init(cfg.canvas.size, cfg.font.color);
      canvas = document.getElementById(cfg.canvas.id);
      ctx = canvas.getContext('2d');
      font = cfg.font.size + 'px ' + cfg.font.name;
      bold = 'bold ' + font;
      ctx.font = font;
      dx = ctx.measureText(calibration).width / calibration.length;
      dy = cfg.font.size + 2;
      canvas.width = cols * dx;
      canvas.height = rows * dy;
      ctx.fillStyle = cfg.font.color.b;
      ctx.fillRect(0, 0, canvas.width, canvas.height);
      cursor.init(cfg.cursor);
      document.addEventListener('keydown', function(evt)
      {
         tty.keyboard.onkeydown(evt);
      }, true);
      document.addEventListener('keypress', function(evt)
      {
         tty.keyboard.onkeypress(evt);
      }, true);
      document.addEventListener('keyup', function(evt)
      {
         tty.keyboard.onkeyup(evt);
      }, true);
      if (cfg.canvas.mouse)
      {
         // mouse click
         canvas.addEventListener('click', function(evt)
         {
            var rect = canvas.getBoundingClientRect();           
            var x = evt.clientX - rect.left;
            var y = evt.clientY - rect.top;
            var c = (x / dx) | 0;
            var r = (y / dy) | 0;
            if ((r >= 0 && r < rows) && (c >= 0 && c < cols))
            {
               cursor.setPosition(r, c);
            }            
         }, true);
      }
      my.cy = tty.isFirefox ? 1 : 0;
      canvas.focus();      
   };
   
   /* set cursor position */   
   my.setCursorPosition = function(x, y)
   {
      cursor.setPosition(x, y);
   };
   
   /* set cursor visibility */   
   my.setCursorStatus = function(visible)
   {
      cursor.visible = visible;
      cursor.show();
   };

   /* clear screen */   
   my.clear = function()
   {
      video.clear(rows, cols, font);
      ctx.fillStyle = font.color.b;
      ctx.fillRect(0, 0, canvas.width, canvas.height);
      cursor.setPosition(0, 0);        
   };
  
   /** get colors from color pair */ 
   var getColors = function(k)
   {       
      var v = {};
      
      if(k === 0)
      {
         v.f = video.color.f;
         v.b = video.color.b;
      }
      else
      {
         // check for colors
         var fg = ((k >>> 4) & 0x07) - 1;
         var bg = (k & 0x07) - 1;
      
         v.f = my.palette[fg];
         v.b = my.palette[bg];
      }
      return v;
   };
   
   /* draw line */   
   my.drawLine = function(line)
   {
      var off = 1;
      while(off < line.length)
      {
         var len = line[off];
         my.drawLineAt(line, off, len);
         off += len;
      }
   };

   /* draw line */   
   my.drawLineAt = function(line, off, len)
   {
      var row = line[off+1];
      var col = line[off+2];
      var a = line[off+3];
      var k = line[off+4];
      var x = col * dx;
      var y = row * dy;
      var text = '';    
      var j = col;
      
      // build string
      for(var i = off+5;i < off+len; i+=2)
      {
         var c = String.fromCharCode(line[i] << 8 | line[i+1]);
         video.put(c, row, j++, a, k);
         text += c;
      }

      // colors     
      var v = getColors(k);
      
      /* draw the text */
      drawText(x, y, text, a, v);       
   };

   /* draw text on canvas*/   
   var drawText = function(x, y, text, a, v)
   {
      var l = text.length;
      var max = dx * l;
      
      ctx.font = (a & 0x02) ? bold : font;                 
      ctx.textBaseline = 'top';
       
      /* reverse */
      if (a & 0x08)
      {
        ctx.fillStyle = v.f;
        ctx.fillRect(x, y, max, dy);      
        ctx.fillStyle = v.b;
        ctx.fillText(text, x, y + my.cy, max);               
        /* underline */
        if (a & 0x04) 
        {
           ctx.fillRect(x, y + dy - 1, max, 1);
        }   
      }
      else
      {
        ctx.fillStyle = v.b;
        ctx.fillRect(x, y, max, dy);      
        ctx.fillStyle = v.f;
        ctx.fillText(text, x, y + my.cy, max);               
        /* underline */
        if (a & 0x04) 
        {
           ctx.fillRect(x, y + dy - 1, max, 1);
        }   
      }
   };

   /* exports */
   return my;       
}());
